#include <iostream.h>
#include <fstream.h>
#include <LEDA/ugraph.h>
#include <LEDA/graph_edit.h>
#include <Tim/graph.h>
#include <Tim/adjacency_matrix.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include "lbfs.h"
#include "interval.h"

char main_version[] = "$Id: main.cc,v 1.13 1996/09/12 18:46:24 tbm Exp $";

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

options_t options;
statistics_t statistics;

/*****************************************************************************/

void statistics_t::count (node v, bool result) {
	num_orders[v]++;
	last_result = result;
	if (result) {
		num_umbrella_orders[v]++;
	}
	if (!found_result[result]) {
		found_result[result] = true;
		if (result) {
			reason_for_interest.append (string ("First umbrella order"));
		} else {
			reason_for_interest.append (string ("First non-umbrella order"));
		}
		result_listed = true;
	}
}

/*****************************************************************************/

void statistics_t::count (const labelled_graph &G, node alpha, flies_status alpha_status, node beta, flies_status beta_status) {
	if (!tb_usage[alpha_status][beta_status]) {
		string reason = "Alpha=";
		reason += G[alpha].label;
		reason += string (" %s and beta=", description (alpha_status));
		reason += G[beta].label;
		reason += string (" %s", description (beta_status));
		reason_for_interest.append (reason);
	}
	tb_usage[alpha_status][beta_status]++;
}

/*****************************************************************************/

void statistics_t::count (node v, const vertex_order &sweep1, const vertex_order &sweep2) {
	if (!(sweep1 == sweep2)) {
		if (!num_mismatches[v]) {
			string reason = "Sweeps 3 and 5 give different orders";
			reason_for_interest.append (reason);
		}
		num_mismatches[v]++;
	}
}

/*****************************************************************************/

void statistics_t::print_reason (ostream &out) {
	if (!interesting ()) {
		return;
	}
	while (!reason_for_interest.empty ()) {
		out << "\t\t" << reason_for_interest.pop () << "\n";
	}
	if (!result_listed) {
		if (last_result) {
			out << "\t\t(Another umbrella order)\n";
		} else {
			out << "\t\t(Another non-umbrella order)\n";
		}
	}
	result_listed = false;
}

/*****************************************************************************/

ostream & operator << (ostream &out, const statistics_t &stats) {
	if (stats.the_graph == NULL) {
		return out;
	}

	out << "\t"       "Number\t"  "Number\t""Number";
	if (options.algorithm == five_sweep) {
		out << "\tNumber";
	}
	out << "\nStart\t"  "initial\t" "umb.\t" "non-umb";
	if (options.algorithm == five_sweep) {
		out << "\tsweep";
	}
	out << "\nvertex\t" "orders\t"  "orders\t""orders";
	if (options.algorithm == five_sweep) {
		out << "\tmismatches";
	}
	out << "\n";

	unsigned long num_orders = 0;
	unsigned long num_umbrella_orders = 0;
	unsigned long num_mismatches = 0;
	node v;
	forall_nodes (v, *stats.the_graph) {
		num_orders += stats.num_orders[v];
		num_umbrella_orders += stats.num_umbrella_orders[v];
		num_mismatches += stats.num_mismatches[v];

		out << (*stats.the_graph)[v].label << "\t";
		out << stats.num_orders[v] << "\t";
		out << stats.num_umbrella_orders[v] << "\t";
		out << stats.num_orders[v] - stats.num_umbrella_orders[v];
		if (options.algorithm == five_sweep) {
			out << "\t" << stats.num_mismatches[v];
		}
		out << "\n";
	}
	out << "Total\t";
	out << num_orders << "\t";
	out << num_umbrella_orders << "\t";
	out << num_orders - num_umbrella_orders;
	if (options.algorithm == five_sweep) {
		out << "\t" << num_mismatches;
	}
	out << "\n";

	out << "\n";
	out << "Tiebreaker Hits:\n";
	out << "\t\t\tBETA\n";
	out << "\t\tFlies\tNFlies\tOK\n";
	out << "\tFlies\t"
		<< stats.tb_usage[FLIES_FLIES][FLIES_FLIES] << "\t"
		<< stats.tb_usage[FLIES_FLIES][FLIES_NEIGHBOUR] << "\t"
		<< stats.tb_usage[FLIES_FLIES][FLIES_OK] << "\n";
	out << "ALPHA\tNFlies\t"
		<< stats.tb_usage[FLIES_NEIGHBOUR][FLIES_FLIES] << "\t"
		<< stats.tb_usage[FLIES_NEIGHBOUR][FLIES_NEIGHBOUR] << "\t"
		<< stats.tb_usage[FLIES_NEIGHBOUR][FLIES_OK] << "\n";
	out << "\tOK\t"
		<< stats.tb_usage[FLIES_OK][FLIES_FLIES] << "\t"
		<< stats.tb_usage[FLIES_OK][FLIES_NEIGHBOUR] << "\t"
		<< stats.tb_usage[FLIES_OK][FLIES_OK] << "\n";

	return out;
}

/*****************************************************************************/

//@Man:
class backtrack_t {
private:
	const labelled_graph *the_graph;
	array<int> next_choice;
	array<int> num_choices;
	int choice;

public:
	/// Reset		\hfill $O(n)$
	inline void reset () {
		choice = -1;
		for (int i = next_choice.low (); i <= next_choice.high(); i++) {
			next_choice[i] = 1;
		}
	}
	/// Rerun last order	\hfill $O(n)$
	inline void rerun () {
		choice = -1;
	}
	/// Set up next order	\hfill $O(n)$
	inline bool done () {
		for (int i = next_choice.high (); i >= next_choice.low (); i--) {
			if ((i <= choice) && (next_choice[i] < num_choices[i])) {
				next_choice[i]++;
				choice = -1;
				return false;
			} else {
				next_choice[i] = 1;
			}
		}
		return true;
	}
	/// Constructor		\hfill $O(n)$
	inline backtrack_t (const labelled_graph &G) : next_choice (G.number_of_nodes ()), num_choices (G.number_of_nodes ()) {
		the_graph = &G;
		reset ();
	};
	/// Tiebreaker		\hfill $O(n)$
	friend tiebreaker backtrack_tiebreaker;
};

/*****************************************************************************/

//@Man:
node backtrack_tiebreaker (
	const labelled_graph &G,
	const lexicographic_order &order,
	const list<node> &choices,
	void *info1,
	void *info2
) {
	backtrack_t *bt = (backtrack_t*) info1;

	if (choices.size () <= 1) {
		return choices.head ();
	}

	bt->choice++;
	bt->num_choices [bt->choice] = choices.size ();

	node v;
	int i = 1;
	forall (v, choices) {
		if (i >= bt->next_choice[bt->choice]) {
			break;
		}
		i++;
	}

	return v;
}

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

void do_stuff (labelled_graph &G, string name, string start_vertex) {
	backtrack_t backtrack (G);

	statistics.init (G);

	time_t clock = time (NULL);
	cout << "Date:      " << ctime (&clock);

	cout << "Versions:  " << interval_version << "\n";
	cout << "           " << lbfs_version << "\n";
	cout << "           " << main_version << "\n";

	cout << "Graph:     " << name << "\n";

	cout << "Algorithm: ";
	if (options.algorithm == old) {
		cout << "old";
	} else {
		if (options.algorithm == three_sweep) {
			cout << "three-sweep (LBFS, LBFS+, LBFS*)";
		} else if (options.algorithm == four_plus) {
			cout << "four-plus (LBFS, LBFS+, LBFS+, LBFS*)";
		} else if (options.algorithm == four_star) {
			cout << "four-star (LBFS, LBFS+, LBFS*, LBFS*)";
		} else {
			cout << "five-sweep (LBFS, LBFS+, LBFS*, LBFS*, LBFS*)";
		}

		cout << "\nFlies:     " << options.flies_name << "\n\n";

		cout << "\t\t\tBETA\n";
		cout << "\t\tFlies\tNFlies\tOK\n";
		cout << "\tFlies\t"
			<< options.flies[FLIES_FLIES][FLIES_FLIES] << "\t"
			<< options.flies[FLIES_FLIES][FLIES_NEIGHBOUR] << "\t"
			<< options.flies[FLIES_FLIES][FLIES_OK] << "\n";
		cout << "ALPHA\tNFlies\t"
			<< options.flies[FLIES_NEIGHBOUR][FLIES_FLIES] << "\t"
			<< options.flies[FLIES_NEIGHBOUR][FLIES_NEIGHBOUR] << "\t"
			<< options.flies[FLIES_NEIGHBOUR][FLIES_OK] << "\n";
		cout << "\tOK\t"
			<< options.flies[FLIES_OK][FLIES_FLIES] << "\t"
			<< options.flies[FLIES_OK][FLIES_NEIGHBOUR] << "\t"
			<< options.flies[FLIES_OK][FLIES_OK];
	}
	cout << "\n";

	adjacency_matrix A(G);
	cout << "\n";
	// cout << A;

	node v;
	int i = A.low1 ();
	forall_nodes (v, G) {
		cout << G[v].label << "\t";
		for (int j = A.low2 (); j <= A.high2 (); j++) {
			if (i == j) {
				cout << "-";
			} else {
				cout << A(i,j);
			}
		}
		cout << "\n";
		i++;
	}

	bool result;
	forall_nodes (v, G) {
                if ((start_vertex.length () > 0) && (start_vertex != G[v].label)) {
			continue;
		}
		backtrack.reset ();
		statistics.reset ();
		do {
			result = (*options.algorithm) (G, v, backtrack_tiebreaker, &backtrack);
			statistics.count (v, result);
			if (options.algorithm == five_sweep) {
				statistics.count (v, sweep[2], sweep[4]);
			}
			if (statistics.interesting ()) {
				cout << "\n";
				for (i = 0; i < options.num_sweeps; i++) {
					cout << "Sweep " << (i+1) << ":\t";
					cout << sweep[i] << "\n";
				}
				statistics.print_reason (cout);
				if (options.verbose) {
					options.full_info = true;
					backtrack.rerun ();
					(*options.algorithm) (G, v, backtrack_tiebreaker, &backtrack);
					options.full_info = false;
				}
				cout.flush ();
			}
		} while (!backtrack.done ());
	}

	cout << "\n" << statistics;

	if (options.debug) {
		// LEDA memory usage statistics
		cout << "\nMemory usage:\n";
		print_statistics ();
	}
}

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

void usage (char *command) {
	cout << "Usage: " << command << " [options] file...\n";
	cout << "\tFiles may be in graphed format (must end in .g) or\n";
	cout << "\tin LEDA format (must end in .ggg).\n";
	cout << "Options:\n";
	cout << "   -h         Display this help information\n";
	cout << "   -v         Verbose\n";
	cout << "   -d         Debug\n";
	cout << "   -e         Edit graph (can only save in LEDA format)\n";
	cout << "   -f [strong|weak|mixed]\n";
	cout << "              Use strong/weak/mixed definition of flies (default is mixed)\n";
	cout << "   -a [old|three-sweep|four-plus|four-star|five-sweep]\n";
	cout << "              Algorithm to use (default is three-sweep)\n";
	cout << "   -s vertex  Specify a starting vertex\n";
}

/*****************************************************************************/
/*                                                                           */
/*****************************************************************************/

int main (int argc, char *argv[]) {
	int c;
	labelled_graph G;
	string start_vertex;

	for (;;) {
		if ((c = getopt (argc, argv, "a:def:hvs:")) < 0) {
			break;
		}
		switch (c) {
		case 'a':
			if (!strcmp (optarg, "old")) {
				options.algorithm = old;
				options.num_sweeps = 4;
			} else if (!strcmp (optarg, "three-sweep")) {
				options.algorithm = three_sweep;
				options.num_sweeps = 3;
			} else if (!strcmp (optarg, "four-plus")) {
				options.algorithm = four_plus;
				options.num_sweeps = 4;
			} else if (!strcmp (optarg, "four-star")) {
				options.algorithm = four_star;
				options.num_sweeps = 4;
			} else if (!strcmp (optarg, "five-sweep")) {
				options.algorithm = five_sweep;
				options.num_sweeps = 5;
			} else {
				cerr << "Algorithm must be 'old', 'three-sweep', 'four-plus', 'four-star' or 'five-sweep'\n";
				return 1;
			}
			break;
		case 'd':
			options.debug = true;
			break;
		case 'e':
			options.edit = true;
			break;
		case 'f':
			if (!strcmp (optarg, "strong")) {
				options.flies = flies_strong;
				options.flies_name = "Strong";
			} else if (!strcmp (optarg, "weak")) {
				options.flies = flies_weak;
				options.flies_name = "Weak";
			} else if (!strcmp (optarg, "mixed")) {
				options.flies = flies_mixed;
				options.flies_name = "Mixed";
			} else {
				cerr << "Flies definition must be 'strong', 'weak' or 'mixed'\n";
				return 1;
			}
			break;
		case 'h':
			usage (argv[0]);
			return 0;
		case 'v':
			options.verbose = true;
			break;
		case 's':
			start_vertex = string (optarg);
			break;
		case '?':
			usage (argv[0]);
			return 1;
		}
	}

	for (; optind < argc; optind++) {
		string filename (argv[optind]);

		if (filename.tail (2) == ".g") {
			G.read_GraphEd (argv[optind]);
			G.build_labels ();
		} else if (filename.tail (4) == ".ggg") {
			if ((G.read (filename) != 0) && (!options.edit)) {
				cerr << filename << ": error reading graph\n";
				return 1;
			}
		} else {
			if (!options.edit) {
				cerr << filename << ": unknown file type\n";
				return 1;
			}
		}

		if (options.edit) {
			G.edit (filename);
		} else {
			G.make_undirected ();
			do_stuff (G, filename, start_vertex);
		}
	}

	return 0;
}
